/*
   This is a better way to define the generic Bag abstract data type (ADT).
   First define a generic abstract base class that specifies the interface
   to the ADT. Then define (possibly several) implementations of the
   interface.

   This class specifies WHAT a generic Bag should be able to do but does
   not contain any implementation details (that is, there is no HOW in
   this file). In particular this generic abstract class does not specify
   the actual data type that will be stored in a Bag and it does not specify
   how the data will be stored (that detail is left to each implementation).
*/
import java.util.Iterator;

/**
* A Bag is a collection of elements.
* The same element may appear multiple times in a bag.
*/
public abstract class Bag<T> implements Iterable<T>
{
   /**
   * Add new elements to this bag.
   * @param elements
   *   (a variable-arity argument)
   *   one or more new elements that are being inserted
   * @postcondition
   *   A new copy of the element has been added to this bag.
   */
   public abstract void add(T... elements);


   /**
   * Add the contents of another bag to this bag.
   * @param addend
   *   a bag whose contents will be added to this bag
   * @precondition
   *   The parameter, addend, is not null.
   * @postcondition
   *   The elements from addend have been added to this bag.
   * @exception NullPointerException
   *   Indicates that addend is null.
   */
   public abstract void add(Bag<T> addend);


   /**
   * Accessor method to count the number of occurrences of a particular element
   * in this bag.
   * @param target
   *   the element that needs to be counted
   * @return
   *   the number of times that target occurs in this bag
   */
   public abstract int countOccurrences(T target);


   /**
   * Remove one copy of a specified element from this bag.
   * @param target
   *   the element to remove from the bag
   * @return
   *   true if the target element was in this bag, false otherwise
   * @postcondition
   *   If target was found in the bag, then one copy of
   *   target has been removed and the method returns true.
   *   Otherwise the bag remains unchanged and the method returns false.
   */
   public abstract boolean remove(T target);


   /**
   * Determine the number of elements in this bag.
   * @param - none
   * @return
   *   the number of elements in this bag
   */
   public abstract int size( );


   /**
   * Create a new bag that contains all the elements from this bag and the other bag.
   * @param b2
   *   the second bag
   * @precondition
   *   b2 is not null.
   * @return
   *   the union of this bag and the bag b2
   * @exception NullPointerException
   *   Indicates that the argument b2 is null.
   */
   public abstract Bag<T> union(Bag<T> b2);


   /**
   * Create an array representation of the contents of this bag.
   * @return
   *   a T array containing every element of this bag
   */
   public abstract T[] toArray();


   /**
   * Create a String representation of the contents of this bag.
   * @return
   *   a String representation of the contents of this bag
   */
   public abstract String toString();


   /**
   *  Provide an abstract "implementation" of the Iterable interface.
   */
   public abstract Iterator<T> iterator();
}
