/*
   This is an implementation of the generic Bag abstract data type (ADT).
   This class provides all the details of HOW a generic Bag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as a linked list).
*/
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
* An LinkedBag<T> is a collection of elements of type T.
* The same element may appear multiple times in a bag.
*
* @note
*   (1) Beyond Int.MAX_VALUE elements, countOccurrences,
*   size, and grab are wrong.
*   <p>
*   (2) Because of the slow linear algorithms of this class, large bags will have
*   poor performance.
*
*/
public class LinkedBag<T> extends Bag<T>
{
   // Invariant of the LinkedBag<T> class:
   //   1. The elements in the bag are stored on a linked list.
   //   2. The head reference of the list is in the instance variable
   //      head.
   //   3. The total number of elements in the list is in the instance
   //      variable manyNodes.
   private Node<T> head;
   private int manyNodes;


   /**
   * Initialize an empty bag.
   * @param - none
   * @postcondition
   *   This bag is empty.
   **/
   public LinkedBag( )
   {
      head = null;
      manyNodes = 0;
   }


   /**
   * Generate a copy of the given bag (as a LinkedBag<T>).
   * @param  original
   *     Bag<T> to be copied. Subsequent changes to the copy
   *     will not affect this original, nor vice versa.
   **/
   public LinkedBag(Bag<T> original)
   {
      this.head = null;
      this.manyNodes = 0;
      this.add( original );
   }


   /**
   * {@inheritDoc}
   */
   public void add(T... elements)
   {
      for (T i : elements)
      {
         head = new Node<T>(i, head);
         manyNodes++;
      }
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(T target)
   {
      boolean answer = false;
      for (Node<T> cursor = head; cursor != null; cursor = cursor.getLink())
      {
         if ( target.equals(cursor.getData()) )
         {
            cursor.setData( head.getData() ); // copy the head node's data into this node
            head = head.getLink();            // remove the head node
            manyNodes--;
            answer = true;
            break;
         }
      }
      return answer;
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyNodes;
   }


   /**
   * {@inheritDoc}
   */
   public Bag<T> union(Bag<T> b2)
   {
      // use the copy constructor
      LinkedBag<T> result = new LinkedBag<T>(this);

      // use add(Bag<T>)
      result.add(b2);

      return result;
   }


   /**
   *  Provide an implementation of the Iterable<T> interface.
   */
   public Iterator<T> iterator()
   {
      return new LinkedBagIterator( );
   }


   /**
      This is an implementation of the Iterator interface
      that knows how to iterator over a LinkedBag.

      Since the LinkedBag class does not expose any information
      about its internal implementation, this Iterator must
      be a nested class within the LinkedBag class so that the
      iterator object has access to the internals of a LinkedBag
      object.
   */
   private class LinkedBagIterator implements Iterator<T>
   {
      Node<T> cursor;

      public LinkedBagIterator( )
      {
         // Here is where we make use of the fact that
         // this iterator is nested inside of the LinkedBag
         // class. We have access to the private field head.
         this.cursor = head;
      }

      public boolean hasNext()
      {
         return (this.cursor != null);
      }

      public T next()
      {
         if ( hasNext() )
         {
            T temp = this.cursor.getData();
            cursor = this.cursor.getLink();
            return temp;
         }
         else
            throw new NoSuchElementException();
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }//LinkedBagIterator
}
