/*
   This is an implementation of the generic Bag abstract data type (ADT).
   This class provides all the details of HOW a generic Bag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as an array).
*/
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
* An ArrayBag<T> is a collection of elements of type T.
* The same element may appear multiple times in a bag.
*
* @note
*   (1) The capacity of one of these bags can change after it's created, but
*   the maximum capacity is limited by the amount of free memory on the
*   machine. The constructor, addItem, clone,
*   and union will result in an OutOfMemoryError
*   when free memory is exhausted.
*   <p>
*   (2) A bag's capacity cannot exceed the maximum integer 2,147,483,647
*   (Integer.MAX_VALUE). Any attempt to create a larger capacity
*   results in a failure due to an arithmetic overflow.
*   <p>
*   (3) Because of the slow linear algorithms of this
*   class, large bags will have poor performance.
*
* @see ArraysBag
*/
public class ArrayBag<T> extends Bag<T>
{
   // Invariant of the ArrayBag<T> class:
   //   1. The number of elements in the bag is in the instance variable
   //      manyItems, which is no more than data.length.
   //   2. For an empty bag, we do not care what is stored in any of data;
   //      for a non-empty bag, the elements in the bag are stored in data[0]
   //      through data[manyItems-1], and we don't care whats in the
   //      rest of data.
   private T[ ] data;
   private int manyItems;

   /**
   * Initialize an empty bag with an initial capacity of 10.  Note that the
   * addItem method works efficiently (without needing more
   * memory) until this capacity is reached.
   * @param - none
   * @postcondition
   *   This bag is empty and has an initial capacity of 10.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for:
   *   new Object[10].
   */
   public ArrayBag( )
   {
      final int INITIAL_CAPACITY = 10;
      manyItems = 0;
      data = (T[])new Object[INITIAL_CAPACITY];
   }


   /**
   * Initialize an empty bag with a specified initial capacity. Note that the
   * addItem method works efficiently (without needing more
   * memory) until this capacity is reached.
   * @param initialCapacity
   *   the initial capacity of this bag
   * @precondition
   *   initialCapacity is non-negative.
   * @postcondition
   *   This bag is empty and has the given initial capacity.
   * @exception IllegalArgumentException
   *   Indicates that initialCapacity is negative.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[initialCapacity].
   */
   public ArrayBag(int initialCapacity)
   {
      if (initialCapacity < 0)
         throw new IllegalArgumentException
         ("The initialCapacity is negative: " + initialCapacity);
      data = (T[])new Object[initialCapacity];
      manyItems = 0;
   }


   /**
   * Generate a copy of the given bag (as a ArrayBag<T>).
   * @param  original
   *     Bag<T> to be copied. Subsequent changes to the copy
   *     will not affect this original, nor vice versa.
   **/
   public ArrayBag(Bag<? extends T> original)
   {
      data = (T[])new Object[original.size()];
      this.add( original );
   }


   /**
   * Change the current capacity of this bag.
   * @param minimumCapacity
   *   the new capacity for this bag
   * @postcondition
   *   This bag's capacity has been changed to at least minimumCapacity.
   *   If the capacity was already at or greater than minimumCapacity,
   *   then the capacity is left unchanged.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new int[minimumCapacity].
   */
   public void ensureCapacity(int minimumCapacity)
   {
      T[ ] biggerArray;

      if (data.length < minimumCapacity)
      {
         biggerArray = (T[])new Object[minimumCapacity];
         System.arraycopy(data, 0, biggerArray, 0, manyItems);
         data = biggerArray;
      }
   }


   /**
   * Accessor method to get the current capacity of this bag.
   * The add method works efficiently (without needing
   * more memory) until this capacity is reached.
   * @param - none
   * @return
   *   the current capacity of this bag
   */
   public int getCapacity( )
   {
      return data.length;
   }


   /**
   * Reduce the current capacity of this bag to its actual size (i.e., the
   * number of elements it contains).
   * @param - none
   * @postcondition
   *   This bag's capacity has been changed to its current size.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for altering the capacity.
   */
   public void trimToSize( )
   {
      T[ ] trimmedArray;

      if (data.length != manyItems)
      {
         trimmedArray = (T[])new Object[manyItems];
         System.arraycopy(data, 0, trimmedArray, 0, manyItems);
         data = trimmedArray;
      }
   }


   /**
   * {@inheritDoc}
   */
   public void add(T... elements)
   {
      if (manyItems + elements.length > data.length)
      {  // Ensure twice as much space as we need.
         ensureCapacity((manyItems + elements.length)*2);
      }

      System.arraycopy(elements, 0, data, manyItems, elements.length);
      manyItems += elements.length;
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(T target)
   {
      int index; // The location of target in the data array.

      // First, set index to the location of target in the data array,
      // which could be as small as 0 or as large as manyItems - 1.
      // If target is not in the array, then index will be set equal
      // to manyItems;
      for (index = 0; (index < manyItems) && (!target.equals(data[index])); index++)
         /* No work is needed in the body of this for-loop. */
         ;

      if (index == manyItems)
      {  // The target was not found, so nothing is removed.
         return false;
      }
      else
      {  // The target was found at data[index].
         manyItems--;
         // copy the last element onto data[index].
         data[index] = data[manyItems];
         return true;
      }
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyItems;
   }


   /**
   * {@inheritDoc}
   */
   public Bag<T> union(Bag<? extends T> b2)
   {
      // use the copy constructor
      ArrayBag<T> result = new ArrayBag<T>(this);

      // use add(Bag<T>)
      result.add(b2);

      return result;

      // What if b2 == this? (See page 142 of the textbook.)
   }


   /**
   *  Provide an implementation of the Iterable<T> interface.
   */
   public Iterator<T> iterator()
   {
      return new ArrayBagIterator();
   }

   /**
      This is an implementation of the Iterator interface
      that knows how to iterator over an ArrayBag.

      Since the ArraysBag class does not expose any information
      about its internal implementation, this Iterator must
      be a nested class within the ArrayBag class so that the
      iterator object has access to the internals of a ArrayBag
      object.
   */
   private class ArrayBagIterator implements Iterator<T>
   {
      int cursor;

      public ArrayBagIterator( )
      {
         this.cursor = 0;
      }

      public boolean hasNext()
      {
         return ( cursor < manyItems );
      }

      @SuppressWarnings("unchecked")
      public T next()
      {
         // Here is where we make use of the fact that
         // this iterator is nested inside of the ArrayBag
         // class. We have access to the private array
         // data.
         if ( hasNext() )
         {
            return data[cursor++];
         }
         else
            throw new NoSuchElementException();
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }//ArrayBagIterator
}
