/*
   This is an implementation of the generic Bag abstract data type (ADT).
   This class provides all the details of HOW a generic Bag does what
   it does. In particular, this class specifies how the actual data
   will be stored (as two arrays).
*/
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
* An ArraysBag<T> is a collection of elements of type T.
* The same element may appear multiple times in a bag.
*
* @note
*   (1) The capacity of one of these bags can change after it's created, but
*   the maximum capacity is limited by the amount of free memory on the
*   machine. The constructor, addItem, clone,
*   and union will result in an OutOfMemoryError
*   when free memory is exhausted.
*   <p>
*   (2) A bag's capacity cannot exceed the maximum integer 2,147,483,647
*   (Integer.MAX_VALUE). Any attempt to create a larger capacity
*   results in a failure due to an arithmetic overflow.
*   <p>
*   (3) Because of the slow linear algorithms of this
*   class, large bags will have poor performance.
*
* @see ArrayBag
*/
public class ArraysBag<T> extends Bag<T>
{
   // The two arrays data and dataCounts are "parallel arrays"
   // The data[] array contains the elements that are in the bag.
   // The dataCounts[] array  contains the count of the occurrences
   // of each element in the bag.
   // Say data[i] is e and dataCounts[i] is 12.
   // That means the bag holds 12 occurrences of the element e.
   // Note that if data[i] is e, then data[j] cannot be e for any j != i.
   private T[ ] data;
   private int[ ] dataCounts;
   private int manyItems;      // the number of entries in the bag
   private int distinctItems;  // the number of entries in data[]


   /**
   * Initialize an empty bag with an initial capacity of 10 distinct elememts.
   *
   * @param - none
   *
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for:
   *   new Object[10].
   */
   public ArraysBag( )
   {
      final int INITIAL_CAPACITY = 10;
      manyItems = 0;
      distinctItems = 0;
      data       = (T[])new Object[INITIAL_CAPACITY];
      dataCounts = new int[INITIAL_CAPACITY];
   }


   /**
   * Initialize an empty bag with a specified initial capacity for distinct numbers.
   *
   * @param initialCapacity
   *   the initial capacity of distinct elements in this bag
   *
   * @exception IllegalArgumentException
   *   Indicates that initialCapacity is negative.
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new Object[initialCapacity].
   */
   public ArraysBag(int initialCapacity)
   {
      if (initialCapacity < 0)
         throw new IllegalArgumentException
         ("The initialCapacity is negative: " + initialCapacity);
      data       = (T[])new Object[initialCapacity];
      dataCounts = new int[initialCapacity];
      manyItems = 0;
      distinctItems = 0;
   }


   /**
   * Generate a copy of the given bag (as a ArraysBag<T>).
   * @param  original
   *     Bag<T> to be copied. Subsequent changes to the copy
   *     will not affect this original, nor vice versa.
   **/
   public ArraysBag(Bag<? extends T> original)
   {
      data       = (T[])new Object[original.size()];
      dataCounts = new int[original.size()];
      this.add( original );
   }


   /**
   * Change the current capacity of distinct elements in this bag.
   * @param minimumCapacity
   *   the new capacity of distinct elements for this bag
   * @exception OutOfMemoryError
   *   Indicates insufficient memory for: new Object[minimumCapacity].
   */
   public void ensureCapacity(int minimumCapacity)
   {
      T[ ] biggerArray1;
      int[ ] biggerArray2;

      if (data.length < minimumCapacity)
      {
         biggerArray1 = (T[])new Object[minimumCapacity];
         System.arraycopy(data, 0, biggerArray1, 0, distinctItems);
         data = biggerArray1;

         biggerArray2 = new int[minimumCapacity];
         System.arraycopy(dataCounts, 0, biggerArray2, 0, distinctItems);
         dataCounts = biggerArray2;
      }
   }


   /**
   * Accessor method to get the current capacity of distinct elements for this bag.
   * @param - none
   * @return
   *   the current capacity of this bag
   */
   public int getCapacity( )
   {
      return data.length;
   }


   /**
   * {@inheritDoc}
   */
   public void add(T... elements)
   {
      for (T element : elements)
      {
         int index = 0;
         for (; index < distinctItems; index++)
         {
            if ( element.equals(data[index]) )
               break;
         }

         if (index == distinctItems)
         {  // element is not in the bag
            if (distinctItems == data.length)
            {  // Ensure twice as much space as we need.
               ensureCapacity( (distinctItems + 1)*2 );
            }

            data[distinctItems] = element;
            dataCounts[distinctItems] = 1;
            distinctItems += 1;
            manyItems += 1;
         }
         else
         {  // element is already in the bag
            dataCounts[index] += 1;
            manyItems += 1;
         }
      }
   }


   /**
   * {@inheritDoc}
   */
   public boolean remove(T target)
   {
      int index; // The location of target in the data array.

      // First, set index to the location of target in the data array,
      // which could be as small as 0 or as large as distinctItems - 1.
      // If target is not in the array, then index will be set equal
      // to distinctItems;
      for (index = 0; (index < distinctItems) && (! target.equals(data[index])); index++)
         /* No work is needed in the body of this for-loop. */
         ;

      if (index == distinctItems)
      {  // The target was not found, so nothing is removed.
         return false;
      }
      else
      {  // The target was found at data[index].
         manyItems--;
         dataCounts[index]--;
         if ( 0 == dataCounts[index] )
         {
            distinctItems--;
            data[index]       = data[distinctItems];
            dataCounts[index] = dataCounts[distinctItems];
         }
         return true;
      }
   }


   /**
   * {@inheritDoc}
   */
   public int size( )
   {
      return manyItems;
   }


   /**
   * {@inheritDoc}
   */
   public Bag<T> union(Bag<? extends T> b2)
   {
      // use the copy constructor
      ArraysBag<T> result = new ArraysBag<T>(this);

      // use add(Bag<T>)
      result.add(b2);

      return result;

      // What if b2 == this? (See page 142 of the textbook.)
   }


   /**
   *  Provide an implementation of the Iterable interface.
   */
   public Iterator<T> iterator()
   {
      return new ArraysBagIterator();
   }


   /**
      This is an implementation of the Iterator interface
      that knows how to iterator over an ArraysBag.

      Since the ArraysBag class does not expose any information
      about its internal implementation, this Iterator must
      be a nested class within the ArraysBag class so that the
      iterator object has access to the internals of a ArraysBag
      object.
   */
   private class ArraysBagIterator implements Iterator<T>
   {
      int cursor;       // idnex into data and dataCounts arrays
      int cursorCount;  // multiplicity counter for current data element
      int elementCount; // counts how many elements have been returned by the iterator

      public ArraysBagIterator( )
      {
         this.cursor = 0;
         this.cursorCount = 0;
         this.elementCount = 0;
      }

      public boolean hasNext()
      {
         return ( this.elementCount < manyItems );
      }

      @SuppressWarnings("unchecked")
      public T next()
      {
         // Here is where we make use of the fact that
         // this iterator is nested inside of the ArraysBag
         // class. We have access to the private arrays
         // data and dataCounts.
         if ( hasNext() )
         {
            if ( this.cursorCount < dataCounts[this.cursor] )
            {
               // count one more occurrence of the "current" element
               this.cursorCount++;
            }
            else
            {
               // "move" to the next element
               this.cursor++;
               this.cursorCount = 1;
            }
            this.elementCount++;
            return (T)data[cursor];
         }
         else
            throw new NoSuchElementException();
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }//ArraysBagIterator
}
