/**
   A simple generic implementation of the List abstract data type
   that uses an array based implementation with an iterator.

   Here is the public List interface.
      public int size()
      public boolean isEmpty()
      public T getFirst()
      public void add(T element)
      public T remove()

   Notice that the public interface to a List does not in any
   way let you step through the list. There is no notion of
   a "current" or a "next" item in the list. The only publicly
   accessible element of the list is the first item.

   So any kind of iterator for this list must somehow get
   access to the internal structure of the list. The only
   way to do this is have the Iterator class for this ArrayList
   class be an internal, nested class, inside of the ArrayList
   class. The code in a nested class has the fields of its
   outer class in scope.
*/
import java.util.Iterator;
import java.util.NoSuchElementException;

public class ArrayList<T> extends List<T>
{
   /**
      This is an implementation of the Iterator interface
      that knows how to iterator over an ArrayList.

      Since the List class does not expose any information
      about its internal implementation, this Iterator must
      be a nested class within the ArrayList class so that the
      iterator object has access to the internals of an ArrayList
      object.
   */
   class ArrayListIterator implements Iterator<T>
   {
      int cursor;


      public ArrayListIterator()
      {
         this.cursor = size - 1;
      }

      public boolean hasNext()
      {
         return (0 < cursor);
      }

      @SuppressWarnings("unchecked")
      public T next()
      {
         // Here is where we make use of the fact that
         // this iterator is nested inside of the ArrayList
         // class. We have access to the private array
         // data.
         if ( hasNext() )
         {
            return (T)data[cursor--];
         }
         else
            throw new NoSuchElementException();
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }//ArrayListIterator


   // A private field in the ArrayList class.
   private Object[] data;
   private int size;
   private final int INITIAL_CAPACITY = 100;

   public ArrayList()
   {
      data = new Object[INITIAL_CAPACITY];
      size = 0;
   }

   public int size()
   {
      return size;
   }

   @SuppressWarnings("unchecked")
   public T getFirst()
   {
      return (T)data[size - 1];
   }

   public void add(T element)
   {
      ensureCapacity();
      this.data[size] = element;
      size++;
   }

   @SuppressWarnings("unchecked")
   public T remove()
   {
      T temp = (T)data[size - 1];
      size--;
      return temp;
   }

   private void ensureCapacity()
   {
      if (size == data.length)
         throw new Error("full array");
   }


   // Provide an implementation of the Iterable interface.
   public Iterator<T> iterator()
   {
      return new ArrayListIterator();
   }

}//ArrayList
