/**
   A simple generic implementation of the List abstract data type
   that uses a linked list implementation with an iterator.

   Here is the public List interface.
      public int size()
      public boolean isEmpty()
      public T getFirst()
      public void add(T element)
      public T remove()

   Notice that the public interface to a List does not in any
   way let you step through the list. There is no notion of
   a "current" or a "next" item in the list. The only publicly
   accessible element of the list is the first item.

   So any kind of iterator for this list must somehow get
   access to the internal structure of the list. The only
   way to do this is have the Iterator class for this LinkedList
   class be an internal, nested class, inside of the LinkedList
   class. The code in a nested class has the fields of its
   outer class in scope.
*/
import java.util.Iterator;
import java.util.NoSuchElementException;

public class LinkedList<T> extends List<T>
{
   /**
      This is a "private" class that is just for use by the List class.
      Since the List class does not want to expose it internal, private
      implementation, it should keep the Node class private.
   */
   class Node<T>
   {
      public T       data;
      public Node<T> next;

      public Node(T data, Node<T> next)
      {
         this.data = data;
         this.next = next;
      }
   }//Node


   /**
      This is an implementation of the Iterator interface
      that knows how to iterator over a LinkedList.

      Since the List class does not expose any information
      about its internal implementation, this Iterator must
      be a nested class within the LinkedList class so that the
      iterator object has access to the internals of a LinkedList
      object.
   */
   class LinkedListIterator implements Iterator<T>
   {
      Node<T> cursor;

      public LinkedListIterator( )
      {
         // Here is where we make use of the fact that
         // this iterator is nested inside of the LinkedList
         // class. We have access to the private field
         // first.
         this.cursor = first;
      }

      public boolean hasNext()
      {
         return (this.cursor != null);
      }

      public T next()
      {
         if ( hasNext() )
         {
            T temp = this.cursor.data;
            this.cursor = this.cursor.next;
            return temp;
         }
         else
            throw new NoSuchElementException();
      }

      public void remove()
      {
         throw new UnsupportedOperationException();
      }
   }//LinkedListIterator


   // A private field in the LinkedList class.
   private Node<T> first;

   public LinkedList()
   {
      first = null;
   }

   public int size()
   {
      Node<T> cursor = first;
      int length = 0;

      while (cursor != null)
      {
         length++;
         cursor = cursor.next;
      }
      return length;
   }

   public T getFirst()
   {
      return first.data;
   }

   public void add(T element)
   {
      Node<T> node = new Node<T>(element, first);
      first = node;
   }

   public T remove()
   {
      T data = first.data;
      first = first.next;
      return data;
   }


   // Provide an implementation of the Iterable interface.
   public Iterator<T> iterator()
   {
      return new LinkedListIterator();
   }

}//LinkedList
