/*
   This file demonstrats a slightly more realistic kind of
   "generic method".

   This file is a "library" of static max() methods.

   The max(a, b) method can only return the maximum of two
   items when there is a way to compare the items to each
   other. So the type of the actual parameters to max() is
   important. The max() method can be generic, but only up
   to a point. It can't "generically" find the max of two
   items that can't be compared.

   But the max() method should not care HOW its two parameters
   compare themselves.
*/
import java.util.Comparator;
import java.awt.Color;

public class Max
{
   /*
      Overloaded max() methods on some primitive types.
   */
   public static int max(int a, int b)
   {
      return (a >= b) ? a : b;
   }

   public static double max(double a, double b)
   {
      return (a >= b) ? a : b;
   }

   public static char max(char a, char b)
   {
      return (a >= b) ? a : b;
   }


   /*
      Overloaded max() methods on some reference types.

      Notice that we compare reference types in
      a way that is different from primitive types.
      We need the reference types to have a compareTo() method.

      https://docs.oracle.com/javase/8/docs/api/java/lang/Comparable.html
   */
   public static String max(String a, String b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }

   public static Integer max(Integer a, Integer b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }

   public static Double max(Double a, Double b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }

   public static Character max(Character a, Character b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }


/*
   // This method doesn't compile because Color objects
   // don't know how to compare themselves.
   // (The Color class does not implement the Comparable interface).
   public static Color max(Color a, Color b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }
*/


   /**
      Here is the modern generic version of max().

      We must specify that this is generic over only
      those classes that know how to compare themselves.
   */
// @SuppressWarnings({"unchecked", "rawtypes"})
   public static <T extends Comparable<T>> T max(T a, T b)
   {
      return (a.compareTo(b) > 0) ? a : b;
   }



   /**
      Here is a different kind of generic version of max().

      This is generic over any class, as long as we are passed
      an object that knows how to do the comparing.
   */
   public static <T> T max(T a, T b, Comparator<T> cmp)
   {
      return (cmp.compare(a, b) > 0) ? a : b;
   }



   /**
     Here is another kind of "polymorphism".
     We can overload max() to take three parameters.
   */
   public static <T extends Comparable<T>> T max(T a, T b, T c)
   {
      return max(a, max(b, c));
   }


   public static <T> T max(T a, T b, T c, Comparator<T> cmp)
   {
      return max(a, max(b, c, cmp), cmp);
   }
}
