/*
    Recall that a method can only return one value. A common problem in
    programming is the need to have a method return more than one value.
    The solution is to put the multiple values in some kind of "container"
    and have the method return a reference to an instance of the container.

    Quite often the container is an array, but that means that each of the
    values in the container must have the same type (say, three ints). What
    if you want to have a method that returns two ints and a String? Then
    you cannot use an array as the container to hold the return values.

    In the examples below, there are three methods that each return multiple
    values in a container. The containers are various kinds of "Boxes". Notice
    how redundant the code is. The definition of the BoxA container class (used
    by methodA() ) is almost exactly the same as the definitions of the BoxB and
    BoxC containers (used by methodB() and methodC(), respectively). It is never
    a good idea for a program to have this much redundant, parallel, code. There
    should be a way to "factor out" this redundant code. A technique for doing
    this kind of factoring is "generic programming", which is demonstrated in
    the next two files, MotivateGenerics2.java and MotivateGenerics3.java.

    Generic programming is far more useful than as just a way to define method
    return types. These three example programs are meant to demonstrate just one
    kind of problem that generic programming can solve for us. In this data structures
    course we will see many other ways in which to use generic programming.
*/

public class MotivateGenerics1
{
   public static BoxA methodA()  // returns two Integers and a String
   {
      return new BoxA( 101, 2002, "a test" ); // uses autoboxing
   }

   public static BoxB methodB()  // returns an Integer, a Double, and a String
   {
      return new BoxB( 101, 2.02, "another test" ); // uses autoboxing
   }

   public static BoxC methodC()  // returns two Booleans and a String
   {
      return new BoxC( false, true, "more testing" ); // uses autoboxing
   }


   static class BoxA  // inner class, used for the return type of methodA()
   {
      private Integer x;
      private Integer y;
      private String  z;

      public BoxA(Integer x, Integer y, String z)
      {
         this.x = x;
         this.y = y;
         this.z = z;
      }

      public Integer getX() { return x; }
      public Integer getY() { return y; }
      public String  getZ() { return z; }
   }//inner class BoxA


   static class BoxB  // inner class, used for the return type of methodB()
   {
      private Integer x;
      private Double  y;
      private String  z;

      public BoxB(Integer x, Double y, String z)
      {
         this.x = x;
         this.y = y;
         this.z = z;
      }

      public Integer getX() { return x; }
      public Double  getY() { return y; }
      public String  getZ() { return z; }
   }//inner class BoxB


   static class BoxC  // inner class, used for the return type of methodC()
   {
      private Boolean x;
      private Boolean y;
      private String  z;

      public BoxC(Boolean x, Boolean y, String z)
      {
         this.x = x;
         this.y = y;
         this.z = z;
      }

      public Boolean getX() { return x; }
      public Boolean getY() { return y; }
      public String  getZ() { return z; }
   }//inner class BoxC


   public static void main(String[] args)
   {
      BoxA box1 = methodA();  // return values stored in a BoxA object
      int i = box1.getX().intValue();
      System.out.println( "the first integer returned by methodA is " + i );

      BoxC box2 = methodC();  // return values stored in a BoxC object
      boolean b = box2.getX().booleanValue();
      System.out.println( "the first boolean returned by methodC is " + b );
   }
}//MotivateGenerics1
