/*
    In this program, there is a single Box class that is used as the
    return type for all three methods, even though each method returns
    different kinds of values. The trick is that the Box class only
    stores Object references, so a Box object can store three references
    to any kind of object. This has the advantage of eliminating the
    redundant code found in the previous BoxA, BoxB and BoxC classes.
    But there are a couple of problems with this technique. One is
    readability. It looks like, from the public interfaces of the three
    methods methodA(), methodB() and methodC(), that they all have the
    same return type. You can only tell that they have different return
    types by looking carefully at the implementations of the methods. Also,
    when we want to use objects stored in a Box object, we need to know
    what the type of the object is and cast the object to that type. This
    is error prone and can lead to runtime casting errors.

    This kind of "generic programming" was used up through Java JDK 1.4.
    In JDK 1.5, true generic classes were introduced (see the next example)
    and now this kind of generic programming is no longer used as much in Java.
*/

public class MotivateGenerics2
{
   public static Box methodA()  // returns two Integers and a String
   {
      return new Box( 101, 2002, "a test" ); // uses autoboxing
   }

   public static Box methodB()  // returns an Integer, a Double, and a String
   {
      return new Box( 101, 2.02, "another test" ); // uses autoboxing
   }

   public static Box methodC()  // returns two Booleans and a String
   {
      return new Box( false, true, "more testing" ); // uses autoboxing
   }

   static class Box   // "generic" inner class,
   {                   // used for the return type of methodA(), methodB(), methodC()
      private Object x;
      private Object y;
      private Object z;

      public Box(Object x, Object y, Object z)
      {
         this.x = x;
         this.y = y;
         this.z = z;
      }

      public Object getX() { return x; }
      public Object getY() { return y; }
      public Object getZ() { return z; }
   }//inner class Box


   public static void main(String[] args)
   {
      Box box1 = methodA();  // return values stored in a (generic) Box object
      int i = ((Integer)box1.getX()).intValue();  // returned object must be cast to an Integer
      System.out.println( "the first integer returned by methodA is " + i );

      Box box2 = methodC();  // return values stored in (generic) Box object
      boolean b = ((Boolean)box2.getX()).booleanValue();  // returned object must be cast to a Boolean
      System.out.println( "the first boolean returned by methodC is " +  b );
   }
}//MotivateGenerics2
